#include <vector>
#include <stdlib.h> /* rand() */
#include <sys/time.h>
#include <set>
#include <map>
#include <sys/stat.h>

#include <boost/random/mersenne_twister.hpp>
#include <boost/random/uniform_int.hpp>
#include <boost/random/variate_generator.hpp>

#define DO_TRACK_PC 0
#define DO_DISASM   0
#define DO_RAMDUMP  0

#define BISQBOT_INTERNAL
#include "bisqbotdefs.hh"
#include "bisqbot.hh"
extern "C" unsigned char RAM[0x800];
extern "C" unsigned char PALRAM[0x20];
extern "C" void FCEU_DispMessage(const char *format, ...)
  __attribute__((format(printf,1,2)));

extern "C" unsigned char RefrRAM[0x800];
unsigned char RefrRAM[0x800];

static unsigned NextChange=0;

static boost::mt19937 rng;

#include "rockman1sound.hh"
static class Rockman1Sound Rockman1Sound;
#include "rockman1sound.cc"

struct BisqBotStateData
{
    /* Put here all variables that you want to be saved
     * with Savestates when the core does rerecording.
     * Reset() is only called when seeking back to frame 0,
     * which is not always the case if you use 2 or more steps.
     */
    int Goal;
    double BonusX, BonusY;
    unsigned NumShoots;
    unsigned MaxShoots;
    unsigned EventFrame;
    bool Region;
};
static class BisqBotState: public BisqBotStateBase, public BisqBotStateData
{
public:
    virtual ~BisqBotState() { }
    
    BisqBotState()
    {
        Reset();
    }
    void Reset()
    {
        Goal = 0;
        NextChange = 0;
        NumShoots = 0;
        BonusX = 0;
        BonusY = 0;
        MaxShoots = 0;
        Region = false;
        //EventFrame = 0;
        EventFrame = 70 + boost::uniform_int<int> (0,10) (rng);
    }
    
    void operator= (const BisqBotState& b)
    {
        (BisqBotStateData&)*this = b;
    }

public:
} state;

const BisqBotStateBase* BisqBotNewSaveState()
{
    return new BisqBotState(state);
}

void BisqBotLoadState(const BisqBotStateBase* b)
{
    state = *(BisqBotState*)b;
}

extern "C" {
void FCEUI_GetIVectors(unsigned short *reset, unsigned short *irq, unsigned short *nmi);
unsigned SavedHScroll,SavedVScroll;
extern unsigned char (__attribute__((regparm(1))) *ARead[0x10000])(unsigned);
}

static void Check_RAMdiffs(unsigned S)
{
    static unsigned prev_S;
    
    static std::map<unsigned/*sp*/,std::string> RAMstates;
    std::string cur_ramstate((char*)RAM, (char*)RAM+0x800);
    
    /* Ignore stack differences */
    for(unsigned a=0x180; a<=0x1FF; ++a) cur_ramstate[a] = 0;
    /* Ignore sprite table */
    for(unsigned a=0x200; a<=0x2FF; ++a) cur_ramstate[a] = 0;
    /* Ignore music player differences */
    for(unsigned a=0x500; a<=0x57D; ++a) cur_ramstate[a] = 0;
    for(unsigned a=0xE0; a<=0xF5; ++a) cur_ramstate[a] = 0;
    
    std::string& prev_ramstate = RAMstates[S];
    
    if(cur_ramstate != prev_ramstate)
    {
        if(S > prev_S) /* returning from function */
        {
            unsigned wid = 0;
            wid += printf("RAM diffs (to previous time when S=%02X): ", S);
            
            for(unsigned a=0; a<0x800
                              && a < prev_ramstate.size()
                              && a < cur_ramstate.size();
                              ++a)
            {
                if(cur_ramstate[a] != prev_ramstate[a])
                {
                    if(wid > 76) { printf("\n  "); wid = 2; }
                    wid += printf("%03X:%02X ", a, (unsigned char)cur_ramstate[a]);
                }
            }
            printf("\n");
        }
        
        prev_ramstate = cur_ramstate;
    }
    prev_S = S;
}

static const struct {const char*pat;unsigned bytes;bool rel;} addrmodes[12] =
{
    {"%.3s"         ,0,0},
    {"%.3s #%s"     ,1,0},
    {"%.3s %s"      ,1,1},
    {"%.3s %s"      ,1,0},
    {"%.3s %s,x"    ,1,0},
    {"%.3s %s,y"    ,1,0},
    {"%.3s (%s,x)"  ,1,0},
    {"%.3s (%s),y"  ,1,0},
    {"%.3s %s"      ,2,0},
    {"%.3s %s,x"    ,2,0},
    {"%.3s %s,y"    ,2,0},
    {"%.3s (%s)"    ,2,0},
};
static const char asminfo[] =
 // addressing modes
 "AGZZZDDZABAZZIIZ" "CHZZZEEZAKZZZJJZ" "IGZZDDDZABAZIIIZ" "CHZZZEEZAKZZZJJZ"
 "AGZZZDDZABAZIIIZ" "CHZZZEEZAKZZZJJZ" "AGZZZDDZABAZLIIZ" "CHZZZEEZAKZZZJJZ"
 "ZGZZDDDZAZAZIIIZ" "CHZZEEFZAKAZZJZZ" "BGBZDDDZABAZIIIZ" "CHZZEEFZAKAZJJKZ"
 "BGZZDDDZABAZIIIZ" "CHZZZEEZAKZZZJJZ" "BGZZDDDZABAZIIIZ" "CHZZZEEZAKZZZJJZ"
 // opcodes
 "brkora" "??????" "???ora" "asl???" "phpora" "asl???" "???ora" "asl???"
 "bplora" "??????" "???ora" "asl???" "clcora" "??????" "???ora" "asl???"
 "jsrand" "??????" "bitand" "rol???" "plpand" "rol???" "bitand" "rol???"
 "bmiand" "??????" "???and" "rol???" "secand" "??????" "???and" "rol???"
 "rtieor" "??????" "???eor" "lsr???" "phaeor" "lsr???" "jmpeor" "lsr???"
 "bvceor" "??????" "???eor" "lsr???" "clieor" "??????" "???eor" "lsr???"
 "rtsadc" "??????" "???adc" "ror???" "plaadc" "ror???" "jmpadc" "ror???"
 "bvsadc" "??????" "???adc" "ror???" "seiadc" "??????" "???adc" "ror???"
 "???sta" "??????" "stysta" "stx???" "dey???" "txa???" "stysta" "stx???"
 "bccsta" "??????" "stysta" "stx???" "tyasta" "txs???" "???sta" "??????"
 "ldylda" "ldx???" "ldylda" "ldx???" "taylda" "tax???" "ldylda" "ldx???"
 "bcslda" "??????" "ldylda" "ldx???" "clvlda" "tsx???" "ldylda" "ldx???"
 "cpycmp" "??????" "cpycmp" "dec???" "inycmp" "dex???" "cpycmp" "dec???"
 "bnecmp" "??????" "???cmp" "dec???" "cldcmp" "??????" "???cmp" "dec???"
 "cpxsbc" "??????" "cpxsbc" "inc???" "inxsbc" "nop???" "cpxsbc" "inc???"
 "beqsbc" "??????" "???sbc" "inc???" "sedsbc" "??????" "???sbc" "inc???";

static void DisAsm(unsigned PC,unsigned A,unsigned X,unsigned Y,unsigned S,unsigned P)
{
    //Check_RAMdiffs(S);
    
    const unsigned char opcode = ARead[PC](PC);
    const unsigned addrmode = asminfo[opcode]-'A';
    const unsigned bytes = 1+addrmodes[addrmode].bytes;
    bool indexed = addrmode==9||addrmode==10;
    printf("%04X: ", PC);
    for(unsigned b=0; b<bytes; ++b) printf("%02X ", ARead[PC+b](PC+b));
    for(unsigned b=bytes; b<5; ++b) printf("   ");
    char OpBuf[512];
    char OpParamStr[50];

    unsigned opparam = 0;
    for(unsigned b=1; b<bytes; ++b) opparam |= (ARead[PC+b](PC+b)) << (8*(b-1));
    if(addrmodes[addrmode].rel) opparam = PC + (signed char)opparam + 2;
    unsigned opbytes = bytes-1; if(addrmodes[addrmode].rel) ++opbytes;
    sprintf(OpParamStr, "$%0*X", opbytes*2, opparam);
    std::string o = OpParamStr;
    std::string oo = indexed ? "Object" : "Megaman";
    
    unsigned addrinfo = opparam;
    if(addrmode >= 3)
    {
        switch(opparam)
        {
#if 0 /* ROCKMAN2 */
            case 0x400: o = oo+"Type"; break;
            case 0x420: o = oo+"Flags";  break;
            case 0x440: o = oo+"PosScreen"; break;
            case 0x460: o = oo+"PosX"; break;
            case 0x480: o = oo+"PosXFraction"; break;
            case 0x4A0: o = oo+"PosY"; break;
            case 0x4C0: o = oo+"PosYFraction"; break;
            case 0x600: o = oo+"SpeedX"; break;
            case 0x620: o = oo+"SpeedXFraction"; break;
            case 0x640: o = oo+"SpeedY"; break;
            case 0x660: o = oo+"SpeedYFraction"; break;
            case 0x6C0: o = oo+"LifeMeter"; break;
            case 0x200: o = "SpriteTable"; break;
            case 0x1D: o = "NMI_GfxUpdateDone"; break;
            case 0x1C: o = "FrameCounter"; break;
            case 0x14: o = "CurrentBeginScreen"; break;
            case 0x15: o = "CurrentEndScreen"; break;
            case 0x38: o = "CurrentOrderNum"; break;
            case 0x4A: o = "RandomSeed"; break;
            case 0x4B: o = "BlinkStatus"; break;
            case 0xB1: o = "BossStrategyNumber"; break;
            case 0xB2: o = "BossStrategyCounter"; break;
            case 0xB3: o = "BossNumber"; break;
            
            case 0xC000: o = "SwitchBank"; break;
            case 0xC07F: o = "WaitNextFrame"; break;
            case 0xCC6C: o = "ClearSpriteTable"; break;
            case 0xCFA8: o = "RenderMeter"; break;
            case 0xCF5D: o = "RenderMeters"; break;
            
            case 0xCEF9: o = "RenderSomething"; break;
            case 0xCC77: o = "RenderEverything"; break;
            
            case 0xD0D7: o = "ReadJoypads"; break;
            
            case 0xEEEF: o = "MoveObject"; break;
#endif
#if 0 /* LEGEND OF KAGE */
            case 0x0106: o = "PPU2000"; break;
            case 0x0107: o = "PPU2001"; break;
            case 0x0109: o = "PPUXScroll"; break;
            case 0x010A: o = "PPUYScroll"; break;
            case 0x010D: o = "PPUWriteBuf_32bytes"; break;
            case 0xA48E: o = "UpdateScrolling"; break;
            case 0xA7BC: o = "UpdateScrolling2"; break;
            case 0xA4E5: o = "UpdateRandom"; break;
            case 0x0F: o = "RndSeed"; break;
            case 0xC6: o = "NMIactiveflag"; break;
            case 0x0101: o = "KeyPadThisFrame"; break;
            case 0x0102: o = "KeyPadPrevFrame"; break;
            case 0x0103: o = "KeyPadThisFrameCopy"; break;
            case 0x0105: o = "NewKeysThisFrame"; break;
            case 0xC7: o = "GamePaused"; break;
            case 0x6A2: o = "Channel0SFXptr_lo"; break;
            case 0x6A3: o = "Channel0SFXptr_lo"; break;
            case 0x6B2: o = "Channel1SFXptr_lo"; break;
            case 0x6B3: o = "Channel1SFXptr_lo"; break;
            case 0x6C2: o = "Channel2SFXptr_lo"; break;
            case 0x6C3: o = "Channel2SFXptr_lo"; break;
            case 0x6D2: o = "Channel3SFXptr_lo"; break;
            case 0x6D3: o = "Channel3SFXptr_lo"; break;
            case 0x6E2: o = "Channel4SFXptr_lo"; break;
            case 0x6E3: o = "Channel4SFXptr_lo"; break;
#endif

        }
        if(addrmode == 4 || addrmode == 9) opparam += X;
        if(addrmode == 5 || addrmode == 10) opparam += Y;
    }
    sprintf(OpBuf, addrmodes[addrmode].pat, &asminfo[256+opcode*3], o.c_str());
    
    o = OpBuf;
    for(unsigned tmp = 0xFF; tmp > S; --tmp) o = " " + o;
    
    char OpInfo[512] = "";
    if(addrmode >= 3 && addrmode != 6 && addrmode != 7
    && addrmode != 11)
    {
        sprintf(OpInfo, " [$%04X]", opparam);
    }
    printf("%-30s %10s A=%02X X=%02X Y=%02X S=%02X P=%02X\n", o.c_str(), OpInfo, A,X,Y, S, P);
}

void BisqBotMonitorCPU
  (unsigned short PC,
   unsigned char A,
   unsigned char X,
   unsigned char Y,
   unsigned char S,
   unsigned char P)
{
    static unsigned short irq=0, nmi=0, res=0;
    if(!irq) { FCEUI_GetIVectors(&res,&irq,&nmi); }

    static unsigned prevpc=0;
    { static unsigned curpc; prevpc=curpc; curpc=PC; }
    
#if DO_TRACK_PC
   #if 1
    static std::map<unsigned, std::set<unsigned> > jptpc;
    if(/*prevpc == 0x83EE
    || prevpc == 0x8276
    || prevpc == 0x8803
    || prevpc == 0x8EBD
    || prevpc == 0xB821
    || prevpc == 0xF243
      */ARead[prevpc](prevpc) == 0x20) // jsr
    {
        if(jptpc[prevpc].find(PC) == jptpc[prevpc].end())
        {
            FCEU_DispMessage("New PC $%X called from $%X\n", PC, prevpc);
            jptpc[prevpc].insert(PC);
        }
    }
   #endif
   #if 0 /* Solomon's Key */
    if(PC == 0xC563) { state.Goal += 1; FCEU_DispMessage("Got KEY\n"); }
    if(PC == 0xC587) { state.Goal += 2; FCEU_DispMessage("Got DOOR\n"); }
   #endif
#endif

#if 0
unsigned addr=PC;
    if(0
|| (addr >= 0x80F0 && addr <= 0x80F2)
|| (addr >= 0x81D8 && addr <= 0x81D8)
|| (addr >= 0x81DA && addr <= 0x81DA)
)
    {
        fprintf(stderr, "PC=%X A=%02X X=%02X Y=%02X prevPC=%04X reset=%04X irq=%04X nmi=%04X\n",
            PC,A,X,Y, prevpc, res,irq,nmi);
       // exit(1);
    }
#endif

#if DO_DISASM /* DISASSEMBLY */
  if(PC >= 0xB425 && PC <= 0xB465)
  {
    static int InSoundCode=0;
    static unsigned UnLoopPtr=0;
    static bool Looping=false;
    if(PC == irq) { printf("IRQ\n"); InSoundCode=0; Looping=false; }
    if(PC == nmi) { printf("NMI\n"); InSoundCode=0; Looping=false; }
    if(PC == res) { printf("RESET\n"); InSoundCode=0; Looping=false; }
    if(PC == UnLoopPtr) { /*printf("Loop break\n");*/ Looping=false; }
   #if 0 /* ROCKMAN2 */
    if(PC == 0xD0AD) { InSoundCode=0; /*printf("exit sound code\n");*/ }
    if(PC == 0xD0BD) { InSoundCode=0; /*printf("exit sound code\n");*/ }
    if(PC == 0xC000) InSoundCode++; //switchbank
    if(PC == 0xD0D7) InSoundCode++; //readjoypads
    if(PC == 0xCFA8) InSoundCode++; //RenderMeter
    if(PC == 0xCF5D) InSoundCode++; //RenderMeters
    if(PC == 0xCEF9) InSoundCode++; //RenderSomething
    if(PC == 0xCC77) InSoundCode++; //RenderEverything
   #endif
   #if 0 /* KAGE */
   // if(PC == 0x8AA5) { InSoundCode=0; /* exit sound code */ }
    if(PC == 0xA48E) InSoundCode++; //horizscroll
    if(PC == 0xA7BC) InSoundCode++; //updategfx
    if(PC == 0xA4E5) InSoundCode++; //updaterandom
   #endif
    if(!InSoundCode && !Looping)
    {
        DisAsm(PC,A,X,Y,S,P);
        if((ARead[PC](PC) & 0x1F) == 0x10 && (ARead[PC+1](PC+1) >= 0xE0))
        {
            unsigned jump_target = PC + (signed char)(ARead[PC+1](PC+1)) + 2;
            bool valid_short_jump = true;
            while(jump_target < PC)
            {
                unsigned char opcode = ARead[jump_target](jump_target);
                const unsigned addrmode = asminfo[opcode]-'A';
                const unsigned bytes = 1+addrmodes[addrmode].bytes;
                /* Don't hide this short loop, if it contains a branch or a call */
                if(opcode == 0x20
                || ((opcode&0x1F)==0x10)
                || opcode==0x4C
                || opcode==0x6C)
                {
                    valid_short_jump=false;
                    break;
                }
                jump_target += bytes;
            }
            
            if(valid_short_jump)
            {
                Looping=true;
                UnLoopPtr = PC+2;
                printf("<Short loop suppressed>\n");
            }
        }
    }
   #if 0 /* ROCKMAN2 */
    if(PC == 0xC021) { printf("<function suppressed>\n"); InSoundCode--; } //switchbank
    if(PC == 0xD0F4) { printf("<function suppressed>\n"); InSoundCode--; } //readjoypads
    if(PC == 0xCFE2) { printf("<function suppressed>\n"); InSoundCode--; } //RenderMeter
    if(PC == 0xCFA7) { printf("<function suppressed>\n"); InSoundCode--; } //RenderMeters
    if(PC == 0xCEF6) { printf("<function suppressed>\n"); InSoundCode--; } //RenderSomething
    if(PC == 0xCD04) { printf("<function suppressed>\n"); InSoundCode--; } //RenderEverything
    if(PC == 0xD0AA) { InSoundCode++; printf("<sound code call suppressed>\n"); }
    if(PC == 0xD0BA) { InSoundCode++; printf("<sound code call suppressed>\n"); }
   #endif
   #if 0 /* KAGE */
   // if(PC == 0x8AA2) { InSoundCode++; printf("<sound code call suppressed>\n"); }
    if(PC == 0xA4DB) { printf("<function suppressed>\n"); InSoundCode--; } //RenderMeter
    if(PC == 0xA879) { printf("<function suppressed>\n"); InSoundCode--; } //updategfx
    if(PC == 0xA4F3) { printf("<function suppressed>\n"); InSoundCode--; } //updaterandom
   #endif
    if(InSoundCode<0)InSoundCode=0;
    fflush(stdout);
    
    /* Randomness in Kage does not happen in NMI. Routine is A4E5
          tmpvar is at [$0F]
          tmpvar = tmpvar * 5 + 17
     */
  }
#endif

    
#if 0 //ROCKMAN1
    static bool sound=false;
    
    if(PC == 0xD54C) sound=true;
    if(PC == 0xD56E) sound=false;
    //if(sound && PC == 0x9000) Rockman1Sound.Update();
    //if(sound && PC == 0x9003) Rockman1Sound.Command(A,Y);
    
    if(!sound && PC == 0xA238)
    {
        FCEU_DispMessage("Damage from enemy %02X (%u points)\n", Y,A);
        state.Goal = -5;
    }
    if(!sound && PC == 0xBEDF)
    {
        FCEU_DispMessage("Enemy %02X hit with weapon %u (%d points left) - rndseed=%02X\n",
            (unsigned)Y, (unsigned)RAM[0x5F], (signed char)A, (unsigned)RAM[0x46]);
        if(A==0)
            switch(RAM[0x46])
            {
                case 0x5F: case 0x60:
                case 0xC3: case 0xC4:
                    break;
                default:
                    state.Goal = -5;
            }
        if(Y == 0x3A) state.Goal = 2; // bubblebot
    }
    if(!sound
    && (PC == 0x9FE8)) // horiz scrolling
    {
        FCEU_DispMessage("Scrolling routine at PC=%04X A=%02X X=%02X Y=%02X\n", PC, A,X,Y);
        state.Goal = 2;
    }
    if(!sound
    && (PC == 0x925D)) // scroll prev room
    {
        FCEU_DispMessage("Scroll previous room at PC=%04X A=%02X X=%02X Y=%02X\n", PC, A,X,Y);
    }
    if(!sound
    && (PC == 0x92A6)) // scroll next room
    {
        FCEU_DispMessage("Scroll next room at PC=%04X A=%02X X=%02X Y=%02X\n", PC, A,X,Y);
        state.Goal = 2;
    }
    if(PC == 0xC912)
    {
        state.Goal += 2;
        FCEU_DispMessage("Boss %u damaged with weapon %u (%d points left)\n", RAM[0xAC], RAM[0x5F], (signed char)A);
    }
    if(PC == 0xC833)
    {
        FCEU_DispMessage("Acquire item 0x%02X\n", A);
        
        //if(RAM[0x46] >= 0x30) state.Goal += 2; else
        //if(A == 0x3F) state.Goal += 2; else
        state.Goal = -5;
    }
    if(!sound
    && (PC == 0xBF25))
    {
        unsigned objnum = RAM[0x2F];
        state.BonusX = RAM[0x460+objnum]*256 + (unsigned short)(RAM[0x480+objnum]*256 + RAM[0x4A0+objnum])/ 256.0;
        state.BonusY = (unsigned short)(RAM[0x600+objnum]*256 + RAM[0x620+objnum])/ 256.0;

        if(Y == 0x3B)
            FCEU_DispMessage("Create no bonus at %.3f,%.3f\n", state.BonusX,state.BonusY);
        else
            FCEU_DispMessage("Create bonus 0x%02X at %.3f,%.3f\n", (unsigned)Y, state.BonusX,state.BonusY);
        if(Y == 0x3F) state.Goal += 2; else state.Goal = -5;
        /**/
        if(state.BonusX >= 6941.9) state.Goal = -5;
        if(state.BonusX >= 6888.0 && state.BonusX < 6905.0) state.Goal = -5;
        /**/
        //if(state.BonusX < 6872.0) state.Goal = -5;
    }
    
    if(PC == 0xD4B1 && !BotFrontActive()) FCEU_DispMessage("LAG detected\n");
#endif
    if(PC == 0xD4AD) std::memcpy(RefrRAM, RAM, 0x800);


#if 0  /* KAGE */
    if(PC == 0x973F) fprintf(stderr, "3000 points at 973F\n");
    if(PC == 0x9765) fprintf(stderr, "3000 points at 9765\n");
    if(PC == 0x9854) fprintf(stderr, "3000 points at 9854\n");
    //if(PC == 0xAF36) fprintf(stderr, "ConsiderNinja, called from %X\n", prevpc);
    if(PC == 0xAF31)
        switch(prevpc)
        {
            case 0xAF11: /*fprintf(stderr, "No enemy generated, music is wrong\n");*/ break;
            case 0xAF1B: /*fprintf(stderr, "No enemy, timer is bad\n");*/break;
            case 0xAF15: /*fprintf(stderr, "No enemy generated, blocked by $85\n");*/ break;
            case 0xAF99: state.Goal=-1; fprintf(stderr, "Not generating hokage, because the stage is not clear\n"); break;
            default: fprintf(stderr, "Not generating enemy, called from %X\n", prevpc);
        }
    if(PC == 0xAFC4)
    {
        //fprintf(stderr, "Generate Enemy %02X (frame=%02X)\n", A, RAM[0x21]);
        //if(A < 0x40) state.Goal = -1;
    }
    //if(PC == 0x93E6) fprintf(stderr, "Considering whether to enable ninjas again\n");
    if(PC == 0x9A61) fprintf(stderr, "Reselecting ninja timer (called from %X)\n", prevpc);
    if(PC == 0x9A72 || PC == 0xB45D
    || PC == 0x8333 || PC == 0x848B
    || PC == 0x97A6 || PC == 0x97AE) fprintf(stderr, "@%X: Ninja Timer Set as %02X (called from %X)\n", PC,
        ARead[PC](PC)==0x86 ? X : A, prevpc);
    //if(PC == 0x9404) fprintf(stderr, "Ninjas enabled again\n");
    if(PC == 0xB410)
    {
        fprintf(stderr, "kill one.. of type %02X\n", A);
        if(A >= 0x40)
        {
            fprintf(stderr, "** HOKAGE KILLED @ x=%02X, seen=%u, cball=%u\n",
                RAM[0x8C+X],
                RAM[0x4A], RAM[0x4B]);
                
            if(0
           // || (RAM[0x4B] < 3 && RAM[0x8C+X] < 0x30)
           // || (RAM[0x21]<0xC3 && RAM[0x4A] >= 10)
           // || (RAM[0x21]<0xE3 && RAM[0x4A] >= 11)*/
            || 1
              )
                state.Goal += 3; // KAGE KILLED
        }
        else
        {
            state.Goal += 1;
            if(RAM[0x2E] != 0x03) state.Goal = -1;
            //if(RAM[0x21] >= 0x9E) state.Goal = -1; // too late
            fprintf(stderr, "goal=%d\n", state.Goal);
        }
        //state.Goal += 1;
    }
    if(PC == 0xAF6C) fprintf(stderr, "Enemy %u witholds a crystal ball!\n", X);
    //if(PC == 0x90E5) fprintf(stderr, "%X called from %X\n", PC, prevpc);
    /*if(PC == nmi
    && (unsigned char)RAM[0x84] <= 8) { state.Goal = 2; }
    */
    if(PC == 0xAC69 && A >= 0xF8)
    {
        if(state.Goal == 0)
        {
            //fprintf(stderr, "Enemy %u disappears\n", X);
            //state.Goal = -1;
        }
        else
        {
            //fprintf(stderr, "Enemy %u disappears, doesn't matter\n", X);
        }
    }
    
    /*
    // KAGE APPEARS
    if(RAM[0x89] >= 0x40 && RAM[0x89] <= 0x7F
    || RAM[0x8A] >= 0x40 && RAM[0x8A] <= 0x7F
    || RAM[0x8B] >= 0x40 && RAM[0x8B] <= 0x7F) state.Goal = 2;
    */
    
    //if(RAM[0x4B] < 3) state.Goal = 2; // Crystal ball found 
    
    /*if(RAM[0x31] & 0x80)
    {
        for(unsigned a=0; a<3; ++a)
            if(RAM[0x98+a] > 0) // this one died
                if(RAM[0x8C+a] > 0xC0) // died at right edge
                    state.Goal = 2;
    }*/
#endif

#if 0 //ROCKMAN2

    // 90C9 = DoScrolling
    if(true
    && (PC == 0x9185)) // vert scrolling
    {
        FCEU_DispMessage("VertScroll at PC=%04X A=%02X X=%02X Y=%02X\n", PC, A,X,Y);
        state.Goal += 0;
    }
    if(true
    && (PC == 0x90D2)) // horiz scrolling
    {
        FCEU_DispMessage("HorizScroll at PC=%04X A=%02X X=%02X Y=%02X\n", PC, A,X,Y);
        state.Goal += 2;
    }
    if(true
    && (PC == 0x8F39)) // next-scrolling
    {
        FCEU_DispMessage("ScrollNextRoom at PC=%04X A=%02X X=%02X Y=%02X -- $37=%u, Rockman Y=%.2f speed %.2f, item1 %02X\n",
            PC, A,X,Y, RAM[0x37],
            (unsigned short)(RAM[0x4A0]*256 + RAM[0x4C0])/256.0,
            (signed short)(RAM[0x640]*256 + RAM[0x660])/256.0,
            RAM[0xA4]);
        if(RAM[0x37] & 1) state.Goal = -5; else state.Goal += 2;
        // vert = bad, horiz = good
    }
    if(PC == 0x80AB)
    {
        FCEU_DispMessage("Boss %u damaged with weapon %u (%d points left)\n", RAM[0xB3], RAM[0xB3], (signed char)A);
    }
    if(PC == 0xF262)
    {
        FCEU_DispMessage("Create no bonus\n");
    }
    if(PC == 0xF27B)
    {
        FCEU_DispMessage("Create bonus 0x%02X\n", A);
        if(A == 0x78) state.Goal += 2; else state.Goal -= 5; 
    }
    
    if(!BotFrontActive())
    {
        if(/*PC == 0xCFFA // megaman2
        || */PC == 0xCFF7 // rockman2
          ) FCEU_DispMessage("LAG detected at PC=%04X\n", PC);
    }
#if 0
    if(RAM[0x6C1] < 28 - 1*6)
    {
        state.Goal += 2;
        if(BotFrontActive())
            FCEU_DispMessage("Boss HP is %u\n", RAM[0x6C1]);
    }
#endif
    
    /*
    static int counter=0;
    if(++counter>=50000) {counter=0;
    fprintf(stderr, "Object types: ");
    for(unsigned a=0; a<32; ++a)
        fprintf(stderr, " %u", RAM[0x400 + a]);
    fprintf(stderr, "\n");}*/
    
    /*if(PC == nmi)
    {
        FCEU_DispMessage("Y=%.2f speed=%.2f\n",
            (unsigned short)(RAM[0x4A0]*256 + RAM[0x4C0])/256.0,
            (signed short)(RAM[0x640]*256 + RAM[0x660])/256.0);
    }*/
#endif

#if 0
    if(false && PC == 0x92B6)
    {
        FCEU_DispMessage("stage=%02X A=%02X CurScreen=%02X\n", RAM[0x31], A, RAM[0x460]);
    }
    if(false && PC == 0xCD8A)
    {
        FCEU_DispMessage("Y=%u 0C=$%02X 0D=$%02X 0E=$%02X\n", Y,RAM[12],RAM[13],RAM[14]);
        fprintf(stderr, "#Actives=%u [%f] (", RAM[0x720], RAM[0x8E]/6.0);
        for(unsigned a=0; a<RAM[0x720]; ++a)
        {
            fprintf(stderr, "%u,%u,%u,%u,%u,%u, ",
                RAM[0x721+a*6],
                RAM[0x722+a*6],
                RAM[0x723+a*6],
                RAM[0x724+a*6],
                RAM[0x725+a*6],
                RAM[0x726+a*6]);
        }
        fprintf(stderr, ")\n");
    }
    //if(PC == 0xCDDE) fprintf(stderr, "CDDE y=%u\n", Y);
#endif
}

void BisqBotReset()
{
    struct timeval tv;
    gettimeofday(&tv,NULL);
    rng.seed(tv.tv_usec + tv.tv_sec*1000000ULL);
    
    state.Reset();
}

template<unsigned mod>
static inline unsigned GetRnd()
{
    static boost::uniform_int<int> g(0, mod-1);
    return g(rng);
}
static inline unsigned GetRnd(unsigned mod)
{
    return boost::uniform_int<int> (0,mod-1) (rng);
}

unsigned char BisqBotInput(unsigned FrameNo)
{
    if(FrameNo >= 97-state.EventFrame) { ++state.MaxShoots; return 0; }

    static const unsigned char InputTable[] =
    {
        K_L,
        K_L,
        K_L,
        K_L,
        K_L | K_A,
        K_L | K_B,
        K_L | K_A | K_B,
        K_L,
        K_L,
        K_L,
        K_L,
        K_L | K_A,
        K_L | K_B,
        K_L | K_A | K_B,
        K_L,
        K_L,
        K_L,
        K_L,
        K_L | K_A,
        K_L | K_B,
        K_L | K_A | K_B,
        K_A,
        K_B,
        K_A,
        K_B,
        K_A,
        K_B,
        K_L,
        K_L,
        0,
        0,
        0,
        0,
        0,
        K_R,
        K_R,
        K_R,
        K_R | K_A,
        K_R | K_B,
        K_R | K_A | K_B,
        K_R | K_U,
        K_L | K_U,
        K_R | K_U | K_B,
        K_L | K_U | K_B,
        K_R | K_U,
        K_L | K_U,
    };
    static unsigned LastInput = 0;
    static unsigned A_State   = 0;
    static unsigned B_State   = 0;
    
    //if(FrameNo >= 26+3) return K_R;
   
    //if(LastInput == 0 && FrameNo < 50) LastInput |= K_A | K_R;
    //if(LastInput & K_L && FrameNo < 50) { LastInput &= ~K_L; LastInput |= K_R; }
    
    if(!NextChange) 
    {
        LastInput = InputTable[ GetRnd<sizeof(InputTable)>() ];
        //if(GetRnd<10>() == 0) LastInput |= K_U;
        
      /*
        LastInput = 0;
        if(GetRnd<10>() == 0) LastInput |= K_A;
        if(GetRnd<5>() == 0) LastInput |= K_B;
        switch(GetRnd<20>())
        {
            //case 0: LastInput |= K_U; break;
            case 1: LastInput |= K_D; break;
        }
        switch(GetRnd<5>())
        {
            case 0: case 1: LastInput |= K_L; break;
            case 2: case 3: LastInput |= K_R; break;
        }
      */
        
        //if(LastInput & K_U)
            NextChange = GetRnd<6>();
        //else
        //    NextChange = GetRnd<15>();
    }
    else --NextChange;
    
#if 0 /* KAGE */
   { unsigned char In = LastInput;
    //if(FrameNo >= 20 && GetRnd<30>() == 0) In = (In ^ (K_L | K_R)) | K_B;
    
    /*if(state.Goal == 0) {
    static bool LastStop=false;
    if(RAM[0x8F] < 0xF0
    && RAM[0x8C] >= 0xF5 && !LastStop) { In &= ~(K_L|K_R); LastStop=true; }
    else LastStop=false;

    }*/
    //if(FrameNo < 90) In &= ~K_B;
    
    return In;}
#endif

#if 0 /* ROCKMAN 1 */
    double RockY  = (unsigned short)(RAM[0x600]*256 + RAM[0x620])/ 256.0;
    double RockX  =
        RAM[0x460]*256 + (unsigned short)(RAM[0x480]*256 + RAM[0x4A0])/ 256.0;
    double RockYS = (signed short)(RAM[0x680]*256 + RAM[0x660])/ 256.0;
#endif
#if 1 /* ROCKMAN 2 */
    double RockY  = (unsigned short)(RAM[0x4A0]*256 + RAM[0x4C0])/ 256.0;
    double RockX  =
        RAM[0x440]*256 + (unsigned short)(RAM[0x460]*256 + RAM[0x480])/ 256.0;
    double RockYS = (signed short)(RAM[0x620]*256 + RAM[0x640])/ 256.0;
#endif
#if 1
    double NextY = RockY-RockYS;
#endif

 //   if(RockX >= 6756 && RockX < 6768) LastInput |= K_A;
 //   if(RockX >= 6740 && RockX < 6756) LastInput &= ~K_A;
    //if(FrameNo < 15) { LastInput &= ~K_L; LastInput |= K_R; }
    //if(FrameNo < 15) { LastInput = K_R | K_A; }
/*
    if(RockY >= 1.5) LastInput = K_R;
*/
    // if(RockX + (26+3-FrameNo) * 1.375 < 6287.0) { state.Goal = -1; }
    
    /*
    if(B_State && FrameNo)
    {
        double RockX = (unsigned short)(RAM[0x480]*256 + RAM[0x4A0])/ 256.0;
        fprintf(stderr,
            "[frame %u]I think the beam went at %u... X=%.1f screen=%u\n",
                FrameNo, RAM[0x5F5], RockX, RAM[0x460]);
    }
    
    B_State = (NextY >= 112 && NextY <= 114) ? K_B : 0;
    if(B_State) fprintf(stderr, "Try beam at RockY=%.1f (%+.1f)\n", RockY, RockYS);
    if(RockY >= 104 && RockY <= 112 && NextY > 113)
        fprintf(stderr, "Rats, this try failed Y=%.1f (%+.1f)\n", RockY, RockYS);
    */
    
    /*
    if(FrameNo < 7) if(GetRnd<3>() == 0) Input |= K_SE;
    if(FrameNo >= 12 && FrameNo < 45) if(GetRnd<20>() == 0) Input |= K_ST;
    */
    
    //if(FrameNo < 16) A_State = K_A; else A_State = 0;
    //if(FrameNo < 43) B_State = K_B; else B_State = 0;
    
    bool Region = false;//RockY >= 78 && RockY < 107;
    if(state.NumShoots < state.MaxShoots && GetRnd<15>() == 0
#if 0
    && (!Region || FrameNo >= 27)
#endif
      )
        { ++state.NumShoots; B_State = K_B;
          if(Region) state.Region=true;
          state.EventFrame = FrameNo;
        }
    else
        B_State = 0;
    
    //if(B_State == K_B) { LastInput &= ~K_L; LastInput |= K_R; }
    //if(FrameNo >= 8 && FrameNo < 15) LastInput |= K_A;
    
    if(GetRnd<15>() == 0) A_State = K_A; else A_State = 0;
    //if(GetRnd<15>() != 0) A_State = K_B; else B_State = 0;
    //if(GetRnd<20>() == 0) Input |= K_U;
    //if(GetRnd<10>() == 0) Input |= K_ST;
    //if(GetRnd<100>() == 0) Input |= K_SE;
    //if(RockY < 1) A_State = 0;

    unsigned char Input = 0x00;
    Input |= LastInput;
    Input |= B_State;
    Input |= A_State;
    
    return Input;
}

extern "C" {
#include <inttypes.h>
extern int current; extern uint32_t framecount; }
void BisqBotRAMreport()
{
   char msg[16384*4] = "";
   extern unsigned char RAM[0x800];
#if 0 /* ROCKMAN 1 */
   double RockY  = (unsigned short)(RAM[0x600]*256 + RAM[0x620])/ 256.0;
   double RockYS = (signed short  )(RAM[0x680]*256 + RAM[0x660])/ 256.0;
   double RockX  = (unsigned short)(RAM[0x480]*256 + RAM[0x4A0])/ 256.0 + RAM[0x460]*256;
   double RockXS = (signed short  )(RAM[0x4C0]*256 + RAM[0x4E0])/ 256.0;
   sprintf(msg, " - X=%.2f%+.4f Y=%.2f%+.4f; random=%02X[%02X,%02X]",
    RockX,RockXS, RockY,RockYS, RAM[0x46],RAM[0x23],RAM[0x0D]);
    
   if(1)
   {
    unsigned char seed=RAM[0x46], frame=RAM[0x23], a;
    #define Iterate(n, c) \
      for(a=0;a<n;++a) { seed ^= c; seed += ++frame; seed >>= 1; }
    Iterate(3, 0x64);
    Iterate(3, 0x7C);
    Iterate(6, 0x64);
    Iterate(1, 0x54);
    Iterate(32, 0x03);
    Iterate(60, 0x02);
    Iterate(1, 0x54); 
    Iterate(33, 0x03);
    Iterate(2, 0x64); 
    Iterate(3, 0x7C); 
    Iterate(6, 0x64); 
    Iterate(1, 0x98); 
    Iterate(13, 0xD4);
    Iterate(1, 0xCC); 
    Iterate(81, 0xD0);
    {char mdata[512];
     sprintf(mdata,",Bomb[%02X]", seed);strcat(msg, mdata);}
   }
   
   if(RockYS <= 0.0)
   {
     strcat(msg, "; jmps=");
     if(RAM[0x620] != 0 && RockYS != -0.75)
     {
       double NewRockY = RockY - RockYS;
       /* Take fraction part only */
       RockY = ((int)RockY) + (NewRockY - (int)NewRockY);
     }
     {double JmpS = 4.8710938;
     unsigned n;for(n=0; n<20; ++n)
     {
         char mdata[512];
         sprintf(mdata, "%d,", (signed int)RockY);
         strcat(msg, mdata);
         RockY -= JmpS;
         JmpS -= 0.25;
         if(RockY < -4) break;
     }}
   }
   
   {unsigned a,b=0;for(a=0; a<16; ++a)
   {
     char mdata[512];
     if(RAM[0x5B0+a]==0) continue;
     
     if(b==0) { ++b; strcat(msg, "\n[36mBeams: "); }
     sprintf(mdata, "%u,%u+%u[%u,%u], ",
        RAM[0x5D0+a]*256 + RAM[0x5C0+a],
        RAM[0x5F0+a],
        RAM[0x5A0+a],
        RAM[0x5E0+a], RAM[0x5B0+a]);
   
     strcat(msg, mdata);
   }}
   strcat(msg, "[37m");
   {unsigned a,b=0;for(a=0; a<32; ++a)
   {
     char mdata[512];
     
     if(RAM[0x600+a] == 0xF8) continue; // dead object
     if(RAM[0x6E0+a] != 0x2C) continue; // lift animator
     
     double LiftX, LiftY;
     if(b==0)
     {
        double LiftX  = (signed short)(RAM[0x9D]*256 + RAM[0x9C])/ 256.0;
        ++b;
        sprintf(mdata, "\n[33mLifts (9B=%02X,9E=%02X,Speed=%.1f): ",
            RAM[0x9B],
            RAM[0x9E],
            LiftX);
        strcat(msg, mdata);
     }

     LiftX  = (unsigned short)(RAM[0x480+a]*256 + RAM[0x4A0+a])/ 256.0 + RAM[0x460+a]*256;
     LiftY  = (unsigned short)(RAM[0x600+a]*256 + RAM[0x620+a])/ 256.0;

     sprintf(mdata, "%.0f,%.0f%s, ", LiftX,LiftY, (RAM[0x640+a]&64) ? "R" : "L");
   
     strcat(msg, mdata);
   }}
   strcat(msg, "[37m");
#endif
#if 0 /* ROCKMAN 2 */
   double RockY  = (unsigned short)(RAM[0x4A0]*256 + RAM[0x4C0])/ 256.0;
   double RockYS = (signed short  )(RAM[0x640]*256 + RAM[0x660])/ 256.0;
   double RockX  = (unsigned short)(RAM[0x460]*256 + RAM[0x480])/ 256.0 + RAM[0x440]*256;
   double RockXS = (signed short  )(RAM[0x600]*256 + RAM[0x620])/ 256.0;
   sprintf(msg, " - X=%.2f%+.3f Y=%.2f%+.3f\n",
    RockX,RockXS, RockY, RockYS
     );
   {
    /*
    sprintf(Buf, "6E0=[%u,%u,%u,%u, %u,%u,%u,%u, %u,%u,%u,%u, %u,%u,%u,%u]",
        RAM[0x6E0], RAM[0x6E1], RAM[0x6E2], RAM[0x6E3],
        RAM[0x6E4], RAM[0x6E5], RAM[0x6E6], RAM[0x6E7],
        RAM[0x6E8], RAM[0x6E9], RAM[0x6EA], RAM[0x6EB],
        RAM[0x6EC], RAM[0x6ED], RAM[0x6EE], RAM[0x6EF]);
    */
    char mdata[512];
    sprintf(mdata, " B0[%u %u %u], Boss[40:%u 42:%u,4E:%u,,68:%u 6A:%u 6C:%u 6E:%u Ypower=%d]",
        RAM[0xB0], RAM[0xB1], RAM[0xB2],
        RAM[0x401],
        RAM[0x421],
    //    RAM[0x441],
    //    RAM[0x461],
    
    //    RAM[0x481],
    //    RAM[0x4A1],
    //    RAM[0x4C1],
        RAM[0x4E1],
        
    //    RAM[0x601],
    //    RAM[0x621],
    //    RAM[0x641],
    //    RAM[0x661],
        
        RAM[0x681],
        RAM[0x6A1],
        RAM[0x6C1],
        RAM[0x6E1],
        
        RAM[0x641]
        );
      strcat(msg, mdata);
   }
#endif
#if 0 /* Battle of olympus */
   // 18 = buttons down this frame (A, B only?)
   // F8 = buttons down this frame
   // F9 = buttons down
   double RockX  = (unsigned short)(RAM[0x10]*256 + RAM[0x52])/256.0;
   double RockY  = (unsigned short)(RAM[0x12]*256 + RAM[0x57])/256.0;
   double RockXS = ((signed char)RAM[0x53]) / 16.0;
   sprintf(msg, " - X=%.2f%+.3f Y=%.2f%+.3f frame=%u\n",
    RockX, RockXS, RockY, 0.0,
    RAM[0x47]);
#endif
#if 0 /* KAGE */
   {unsigned frames_until_enemy = 0;
   for(unsigned n=0; n<256; ++n,++frames_until_enemy)
     if((RAM[0x57] & (unsigned char)(RAM[0x21] + n)) == 0) break;
   sprintf(msg, " - NextEnm@%u(%02X&%02X),SrNo=$%X,Kil=%u,Seen=%u(cball=%u),Remn=$%02X,Hokage=%u, Kage=%02X,%02X, Butterfly=%02X,%02X,6D2=%X EX=%X",
    frames_until_enemy, RAM[0x21],RAM[0x57],
    RAM[0x2E],//screen
    RAM[0x4D],//killed
    RAM[0x4A],//seen
    RAM[0x4B],//crystal ball
    RAM[0x84],//remain
    RAM[0x31],//hokageflag
    
    RAM[0x53],RAM[0x54],
    
    RAM[0xB7],RAM[0xB8],
    
    RAM[0x6D2],
    
    RAM[0x8C]
     );
   };
#endif
#if 0 /* HINOTORI */
   sprintf(msg, " - Hit=%u", RAM[0xD9]);
#endif
#if 0 /* ROCKMAN 4 */
   double RockY  = (unsigned short)(RAM[0x378]*256 + RAM[0x360])/ 256.0;
   double RockYS = (signed short  )(RAM[0x3F0]*256 + RAM[0x3D8])/ 256.0;
   double RockX  = (unsigned short)(RAM[0x330]*256 + RAM[0x318])/ 256.0 + RAM[0x348]*256;
   double RockXS = 0;
   sprintf(msg, " - X=%.2f%+.3f Y=%.2f%+.3f\n",
    RockX,RockXS, RockY, RockYS
     );
#endif
#if 0 /* ROCKMAN 3 */
   double RockY  = (unsigned short)(RAM[0x3C0]*256 + RAM[0x3A0])/ 256.0;
   double RockYS = (signed short  )(RAM[0x460]*256 + RAM[0x440])/ 256.0;
   double RockX  = (unsigned short)(RAM[0x360]*256 + RAM[0x340])/ 256.0 + RAM[0x380]*256;
   double RockXS = 0;
   sprintf(msg, " - X=%.2f%+.3f Y=%.2f%+.3f\n",
    RockX,RockXS, RockY, RockYS
     );
#endif
#if 0 /* GREMLINS 2 */
   #define Load3(addr) \
      ((signed int)((RAM[addr+0]<<24) | (RAM[addr+1]<<16) | (RAM[addr+2]<<8))) / 65536.0
   double RockX  = Load3(0x40);
   double RockXS = Load3(0xA3);
   double RockYS = Load3(0xA0);
   double RockY  = Load3(0x43);
   #undef Load3
   sprintf(msg, " - X=%.2f%+.3f Y=%.2f%+.3f\n",
    RockX,RockXS, RockY, RockYS
     );
#endif
#if DO_RAMDUMP
    /*
        Red = changed smaller
        Green = changed bigger
        Cyan = is not changing anymore
        White = preserved
    */
   {static unsigned char PrevRAM[0x800];
    static char PrevChanges[0x800] = {0};
   {unsigned a;for(a=0x00; a<=0x150; a+=0x01)
   {
     char mdata[512];
     //if(RAM[a] == PrevRAM[a]) continue;
     //if(RAM[a] < 0x4A || RAM[a] > 0x4E) continue;
     //if(a == 0x84) goto Ok;
     /*
     if(a < 0x66 && a != 0x2C && a != 0x2E
                 && a != 0x46 && a != 0x48
                 && a != 0x4C && a != 0x4D
                 && a != 0x56 && a != 0x2B
                 && (a < 0x30 || a > 0x35)) continue;
     if(a > 0x67 && a < 0x6D) continue;
     if(a > 0x6F && a < 0x78) continue;
     if(a > 0x7A && a < 0x86) continue;
     */
     if(/*a >= 0x62 && a <= 0x67 && !((a-0x62)%3)
     || a >= 0x6D && a <= 0x6F && !((a-0x6D)%3)
     || a >= 0x78 && a <= 0x7A && !((a-0x78)%3)
     || a >= 0x86 && a <= 0xB2 && !((a-0x86)%3)
     || ((a < 0x60 || a > 0xB2) &&*/ !(a%16) //)
       )
     {
     Ok:
       sprintf(mdata, "[%04X]", a);
       strcat(msg, mdata);
     }
     if(RAM[a] < PrevRAM[a]) { strcat(msg, "\33[31m"); } // red
     if(RAM[a] > PrevRAM[a]) { strcat(msg, "\33[32m"); } // green
     if(RAM[a] == PrevRAM[a] && PrevChanges[a]) { strcat(msg, "\33[36m"); } // cyan
     sprintf(mdata, "%02X ", RAM[a]);
     strcat(msg, mdata);
     if(RAM[a] != PrevRAM[a] || PrevChanges[a]) { PrevRAM[a]=RAM[a]; strcat(msg, "\33[0m"); }
     PrevChanges[a] = RAM[a] != PrevRAM[a];
   }}
   }
#endif
   FCEU_DispMessage("%s #%u%s",(current<0) ? "Play" : "Rec", framecount, msg);
}

unsigned long long BisqBotEvaluate(unsigned FrameNo)
{
    /*
    #define OkBeam(y) ((y)==113)
    
    state.Goal=1;
    if(OkBeam(RAM[0x5F5]) || OkBeam(RAM[0x5F6])
    || OkBeam(RAM[0x5F7]) || OkBeam(RAM[0x5F8])
    || OkBeam(RAM[0x5F9]))
    {
        if(RAM[0x460] == 5) // screen 5
            state.Goal = 2;
    }*/
    
    /*
    // Fail if screen 5 isn't reached in time
    if(FrameNo >= 40 && RAM[0x460] != 5) { state.Goal=-1; return 0; }
    // Fail if Rockman falls off the beam too soon
    if(FrameNo <= 32 && (signed char)RAM[0x600] > 5) { state.Goal=-1; return 0; }
    if(state.Goal < 0) return 0;
    */
    //if(FrameNo==60) state.Goal = 2;
    
#if 0
    if(BotFrontActive())
    {
        // Fails to fill HP meter?
        if(FrameNo >= 153+3 && RAM[0x6C1] < 1)  {
            fprintf(stderr, "rej, hp 0\n"); state.Goal = -1;
    DBG:
            BisqBotRAMreport();
            return 0; }
        if(RAM[0x6C1] == 28 && RAM[0x641] > 3) {
            fprintf(stderr, "rej, boss acted wrong\n"); state.Goal = -1; goto DBG; }
        if(FrameNo >= 82+3 && RAM[0x440] != 0x19) {
            fprintf(stderr, "rej, not reached room\n"); state.Goal = -1; goto DBG; }
        if(RAM[0x6C1] == 28
        && RAM[0xB1] == 2)
        {
            // Must have at least two frames with B0[3 3 0]
            fprintf(stderr, "1 point\n");
            ++state.Goal;
            
            if(state.Goal == 1) goto DBG;
        }
    }
#endif

#if 0 /* ROCKMAN 1 */
    double RockY  = (unsigned short)(RAM[0x600]*256 + RAM[0x620])/ 256.0;
    double RockX  =
        RAM[0x460]*256 + (unsigned short)(RAM[0x480]*256 + RAM[0x4A0])/ 256.0;
    double RockYS = (signed short)(RAM[0x680]*256 + RAM[0x660])/ 256.0;
#endif
#if 1 /* ROCKMAN 2 */
    double RockY  = (unsigned short)(RAM[0x4A0]*256 + RAM[0x4C0])/ 256.0;
    double RockX  =
        RAM[0x440]*256 + (unsigned short)(RAM[0x460]*256 + RAM[0x480])/ 256.0;
    double RockYS = (signed short)(RAM[0x620]*256 + RAM[0x640])/ 256.0;
#endif
#if 0 /* KAGE */
    //RockY = RAM[0x84]; /* # of enemies remaining */
    RockY = RAM[0x56] + RAM[0x2E]*256;
    
    /*
    if(RAM[0x21] >= 0x20  && RAM[0x21] < 0x60
    && state.Goal == 0) state.Goal = -1;
    */
    
    /*
    if(state.Goal == 1 && !(RAM[0x31] & 0x80))
    {
        // enemy killed but hokage not enabled = bad
        state.Goal = -1;
    }

    if(state.Goal == 0 && RAM[0x8F] < 0xF0 && RAM[0x8C] < 0x80) state.Goal = -1;
    if(state.Goal == 0 && RAM[0x8F] < 0xF0 && RAM[0x8F] < 0x80) state.Goal = -1;
    */
    
    /*
    if(state.Goal == 1
    && (!(RAM[0x31] & 0x80) // hokage was enabled?
     || RAM[0x21] >= 0x20 // it was too late
      )) state.Goal = -1;
    */
    
    //if(FrameNo >= 20 && state.Goal < 1) state.Goal = -1;
    
    //if(GoalFrameNo >= 36 && !(RAM[0x31] & 0x80)) state.Goal = -1;
    
    //if(RAM[0x84] == 0xF8 && frameno >= 200) state.Goal = -1;
#endif

    //if(state.Goal < 1 && RockY >= 117.0) { state.Goal = -5; }
    /*
    if(RockX >= 6848 && RockX < 6868 && RockY + RockYS >= 99) state.Goal = -5;
    */
   // if(RockX >= 1288) state.Goal = -5;
   
#if 0
    if(state.Goal == 0 && RockY < 150) state.Goal=1;
   
    if(RockX < 6436.0 && RockY == 180.0 && state.Goal==1)
    {
        state.Goal = 2;
    }
#endif

    if(state.Goal > 0)
    {
#if 0 /* Shot distances in Rockman 1 */
        double NearestDistance = 256;
        double NearestShot = RockX + 256;
        
        for(unsigned n=2; n<5; ++n)
        {
            if(RAM[0x600+n] == 0xF8) continue; // 
            double ShotX  =
                RAM[0x460+n]*256 + (unsigned short)(RAM[0x480+n]*256 + RAM[0x4A0+n])/ 256.0;
            double Distance = ShotX - RockX;
            if(Distance < 0) Distance = -Distance;
            
            if(Distance < NearestDistance)
                { NearestDistance = Distance; NearestShot = ShotX; }
        }
#endif

        return 2000000000
         //   - FrameNo*256*130
         //   - (unsigned)((RockX-6700)      *256*10)
         //   - (unsigned)((RockY))
            + state.MaxShoots*100
         //   - state.EventFrame*256*10
         //   + (unsigned)(state.BonusX-6700)
         //   + (unsigned)((NearestShot-6700)*256*30)
         ;
        
        //return (10000-FrameNo*3) + RAM[0xA4]*20;
        /*
        double RockY  = (unsigned short)(RAM[0x4A0]*256 + RAM[0x4C0])/ 256.0;
        double YSpeed = (  signed short)(RAM[0x640]*256 + RAM[0x660])/-256.0;
        
        return (10000-FrameNo*3) + (unsigned)(RockY + YSpeed*3); // ok
        */
    }

    return 0;//((RAM[0x460]-4)*256+RAM[0x480]); // not done
}

enum BisqBotInstruction BisqBotTick(unsigned FrameNo, unsigned Step)
{
    if(state.Goal < 0) return BisqBotRestartStep;
#if 0
    if(state.Goal < 2 && FrameNo >= 33 && !state.Region) return BisqBotRestart;
#endif

#if 0
    if(Step == 0 && state.Goal == 1)
    {
        state.MaxShoots = state.NumShoots+1;
        return BisqBotNextStep;
    }
#endif
    if(state.Goal >= 3)
    /*if((signed char)RAM[0x84]
       < (signed char)(1-Step))*/
    //if(RAM[0x31] < 0x80 && RAM[0x31] > Step)
    {
#if 0
            BotFrontMsg("** HOKAGE KILLED @ x=$%02X ($%02X @ screen, kage%+d),y=$%02X seen=%u, cball=%u, TIMER=%02X\n",
                (RAM[0x8C]-RAM[0x56]) + RAM[0x2E]*256,
                RAM[0x8C],
                RAM[0x8C]-RAM[0x53],
                RAM[0x8F],
                RAM[0x4A], RAM[0x4B],
                RAM[0x21]);
#endif

        //double RockY  = (unsigned short)(RAM[0x4A0]*256 + RAM[0x4C0])/ 256.0;
        //BotFrontMsg("RockY=%.2f, BonusX=%.3f, Goal=2\n", RockY, state.BonusX);
        if(Step+1 < MAX_PARTS) return BisqBotNextStep;
        return BisqBotAccept;
    }
    
    // Max item1s used: 5
    //if(RAM[0xA4] < 0x1C - 2*5) return BisqBotRestart;
    
    return BisqBotContinue;
}
